Attribute VB_Name = "Module2"
Option Explicit

Public Const MAXOBJECTS As Long = 3
Public Const MAXDEPTH As Long = 10

Public Type Point
    X As Long
    Y As Long
End Type

Public Type Rect
    Left As Long
    Top As Long
    Right As Long
    Bottom As Long
End Type

Public Type Node
    Center As Point
    Bounds As Rect
    NumPoints As Long
    MyPoints(0 To MAXOBJECTS - 1) As Point
    MyIndex(0 To MAXOBJECTS - 1) As Long
    HasChildren As Boolean
    Parent As Long
    ChildNE As Long
    ChildNW As Long
    ChildSW As Long
    ChildSE As Long
    Level As Long
End Type

Public FoundPoint As Long
Public MinEuclid As Double
Public EuclidsComputed As Long

Public QT1() As Node
Public QT2() As Node
Public CurrentNode As Long

Public Sub InitRoot(ByRef MyBounds As Rect, ByRef Tree() As Node)
    
Tree(0).Bounds.Left = MyBounds.Left
Tree(0).Bounds.Top = MyBounds.Top
Tree(0).Bounds.Right = MyBounds.Right
Tree(0).Bounds.Bottom = MyBounds.Bottom
Tree(0).Center.X = (Tree(0).Bounds.Right - Tree(0).Bounds.Left) \ 2
Tree(0).Center.Y = (Tree(0).Bounds.Bottom - Tree(0).Bounds.Top) \ 2
Tree(0).Parent = -1
Tree(0).HasChildren = True
Tree(0).Level = 0

Tree(0).ChildNE = CreateChild(0, 1, Tree)
Tree(0).ChildNW = CreateChild(0, 2, Tree)
Tree(0).ChildSW = CreateChild(0, 3, Tree)
Tree(0).ChildSE = CreateChild(0, 4, Tree)

End Sub

Public Function CreateChild(ByVal ThisNode As Long, ByVal Side As Long, ByRef Tree() As Node) As Long

'Side:
'1 = NE
'2 = NW
'3 = SW
'4 = SE

CurrentNode = CurrentNode + 1

Tree(CurrentNode).Parent = ThisNode
Tree(CurrentNode).Level = Tree(ThisNode).Level + 1

If Side = 1 Then 'NE
    Tree(CurrentNode).Bounds.Top = Tree(ThisNode).Bounds.Top
    Tree(CurrentNode).Bounds.Left = Tree(ThisNode).Center.X
    Tree(CurrentNode).Bounds.Right = Tree(ThisNode).Bounds.Right
    Tree(CurrentNode).Bounds.Bottom = Tree(ThisNode).Center.Y
End If

If Side = 2 Then 'NW
    Tree(CurrentNode).Bounds.Top = Tree(ThisNode).Bounds.Top
    Tree(CurrentNode).Bounds.Left = Tree(ThisNode).Bounds.Left
    Tree(CurrentNode).Bounds.Right = Tree(ThisNode).Center.X
    Tree(CurrentNode).Bounds.Bottom = Tree(ThisNode).Center.Y
End If

If Side = 3 Then 'SW
    Tree(CurrentNode).Bounds.Top = Tree(ThisNode).Center.Y
    Tree(CurrentNode).Bounds.Left = Tree(ThisNode).Bounds.Left
    Tree(CurrentNode).Bounds.Right = Tree(ThisNode).Center.X
    Tree(CurrentNode).Bounds.Bottom = Tree(ThisNode).Bounds.Bottom
End If

If Side = 4 Then 'SE
    Tree(CurrentNode).Bounds.Top = Tree(ThisNode).Center.Y
    Tree(CurrentNode).Bounds.Left = Tree(ThisNode).Center.X
    Tree(CurrentNode).Bounds.Right = Tree(ThisNode).Bounds.Right
    Tree(CurrentNode).Bounds.Bottom = Tree(ThisNode).Bounds.Bottom
End If

Tree(CurrentNode).Center.X = Tree(CurrentNode).Bounds.Left + ((Tree(CurrentNode).Bounds.Right - Tree(CurrentNode).Bounds.Left) \ 2)
Tree(CurrentNode).Center.Y = Tree(CurrentNode).Bounds.Top + ((Tree(CurrentNode).Bounds.Bottom - Tree(CurrentNode).Bounds.Top) \ 2)

CreateChild = CurrentNode

End Function

Public Sub AddPoint(ByRef ThisPoint As Point, ByVal ThisIndex As Long, ByRef Tree() As Node, Optional ThisNode As Long = 0)

Dim MyChild As Long

'If Tree(ThisNode).Level >= MAXDEPTH Then Stop 'What now? TODO!

If Tree(ThisNode).HasChildren Then 'Child is there. Put it there or iterate deeper. First, select the correct child.
    If ThisPoint.X >= Tree(ThisNode).Center.X And ThisPoint.Y < Tree(ThisNode).Center.Y Then 'NE
        If Not Tree(Tree(ThisNode).ChildNE).HasChildren Then
            If Tree(Tree(ThisNode).ChildNE).NumPoints < MAXOBJECTS Then 'Put it here.
                Tree(Tree(ThisNode).ChildNE).NumPoints = Tree(Tree(ThisNode).ChildNE).NumPoints + 1
                Tree(Tree(ThisNode).ChildNE).MyPoints(Tree(Tree(ThisNode).ChildNE).NumPoints - 1) = ThisPoint
                Tree(Tree(ThisNode).ChildNE).MyIndex(Tree(Tree(ThisNode).ChildNE).NumPoints - 1) = ThisIndex
                Exit Sub
            Else 'Full. Iterate deeper.
                ThisNode = Tree(ThisNode).ChildNE
                Call AddPoint(ThisPoint, ThisIndex, Tree, ThisNode)
            End If
        Else
            ThisNode = Tree(ThisNode).ChildNE
            Call AddPoint(ThisPoint, ThisIndex, Tree, ThisNode)
        End If
    ElseIf ThisPoint.X < Tree(ThisNode).Center.X And ThisPoint.Y < Tree(ThisNode).Center.Y Then 'NW
        If Not Tree(Tree(ThisNode).ChildNW).HasChildren Then
            If Tree(Tree(ThisNode).ChildNW).NumPoints < MAXOBJECTS Then 'Put it here.
                Tree(Tree(ThisNode).ChildNW).NumPoints = Tree(Tree(ThisNode).ChildNW).NumPoints + 1
                Tree(Tree(ThisNode).ChildNW).MyPoints(Tree(Tree(ThisNode).ChildNW).NumPoints - 1) = ThisPoint
                Tree(Tree(ThisNode).ChildNW).MyIndex(Tree(Tree(ThisNode).ChildNW).NumPoints - 1) = ThisIndex
                Exit Sub
            Else 'Full. Iterate deeper.
                ThisNode = Tree(ThisNode).ChildNW
                Call AddPoint(ThisPoint, ThisIndex, Tree, ThisNode)
            End If
        Else
            ThisNode = Tree(ThisNode).ChildNW
            Call AddPoint(ThisPoint, ThisIndex, Tree, ThisNode)
        End If
    ElseIf ThisPoint.X < Tree(ThisNode).Center.X And ThisPoint.Y >= Tree(ThisNode).Center.Y Then 'SW
        If Not Tree(Tree(ThisNode).ChildSW).HasChildren Then
            If Tree(Tree(ThisNode).ChildSW).NumPoints < MAXOBJECTS Then 'Put it here.
                Tree(Tree(ThisNode).ChildSW).NumPoints = Tree(Tree(ThisNode).ChildSW).NumPoints + 1
                Tree(Tree(ThisNode).ChildSW).MyPoints(Tree(Tree(ThisNode).ChildSW).NumPoints - 1) = ThisPoint
                Tree(Tree(ThisNode).ChildSW).MyIndex(Tree(Tree(ThisNode).ChildSW).NumPoints - 1) = ThisIndex
                Exit Sub
            Else 'Full. Iterate deeper.
                ThisNode = Tree(ThisNode).ChildSW
                Call AddPoint(ThisPoint, ThisIndex, Tree, ThisNode)
            End If
        Else
            ThisNode = Tree(ThisNode).ChildSW
            Call AddPoint(ThisPoint, ThisIndex, Tree, ThisNode)
        End If
    ElseIf ThisPoint.X >= Tree(ThisNode).Center.X And ThisPoint.Y >= Tree(ThisNode).Center.Y Then 'SE
        If Not Tree(Tree(ThisNode).ChildSE).HasChildren Then
            If Tree(Tree(ThisNode).ChildSE).NumPoints < MAXOBJECTS Then 'Put it here.
                Tree(Tree(ThisNode).ChildSE).NumPoints = Tree(Tree(ThisNode).ChildSE).NumPoints + 1
                Tree(Tree(ThisNode).ChildSE).MyPoints(Tree(Tree(ThisNode).ChildSE).NumPoints - 1) = ThisPoint
                Tree(Tree(ThisNode).ChildSE).MyIndex(Tree(Tree(ThisNode).ChildSE).NumPoints - 1) = ThisIndex
                Exit Sub
            Else 'Full. Iterate deeper.
                ThisNode = Tree(ThisNode).ChildSE
                Call AddPoint(ThisPoint, ThisIndex, Tree, ThisNode)
            End If
        Else
            ThisNode = Tree(ThisNode).ChildSE
            Call AddPoint(ThisPoint, ThisIndex, Tree, ThisNode)
        End If
    End If
Else 'No children. Check, whether there is room here (then put it here) or, if not, create children, select the right one and put it there.
    If Tree(ThisNode).NumPoints < MAXOBJECTS Then
        Tree(ThisNode).NumPoints = Tree(ThisNode).NumPoints + 1
        Tree(ThisNode).MyPoints(Tree(ThisNode).NumPoints) = ThisPoint
        Tree(ThisNode).MyIndex(Tree(ThisNode).NumPoints) = ThisIndex
        Exit Sub
    ElseIf Tree(ThisNode).Level = MAXDEPTH Then 'If several points are at the exact same spot, the process of creating children
                                                'and moving their content deeper and deeper will loop forever. In those cases, do not add the point
                                                'to the tree. Is this a problem? Not really, as there are points occupying this spot already.
                                                'Closest neighbor search _has_ something to find here.
        Exit Sub
    Else
        Tree(ThisNode).ChildNE = CreateChild(ThisNode, 1, Tree)
        Tree(ThisNode).ChildNW = CreateChild(ThisNode, 2, Tree)
        Tree(ThisNode).ChildSW = CreateChild(ThisNode, 3, Tree)
        Tree(ThisNode).ChildSE = CreateChild(ThisNode, 4, Tree)
        Tree(ThisNode).HasChildren = True
        
        If Tree(ThisNode).NumPoints > 0 Then 'This node now has children, therefore it must not hold points. Remove and pass them deeper.
            Do Until Tree(ThisNode).NumPoints = 0
                Call AddPoint(Tree(ThisNode).MyPoints(Tree(ThisNode).NumPoints - 1), Tree(ThisNode).MyIndex(Tree(ThisNode).NumPoints - 1), Tree)
                Tree(ThisNode).MyPoints(Tree(ThisNode).NumPoints - 1).X = 0
                Tree(ThisNode).MyPoints(Tree(ThisNode).NumPoints - 1).Y = 0
                Tree(ThisNode).MyIndex(Tree(ThisNode).NumPoints - 1) = 0
                Tree(ThisNode).NumPoints = Tree(ThisNode).NumPoints - 1
            Loop
        End If
        
        Call AddPoint(ThisPoint, ThisIndex, Tree, ThisNode) 'Handle initial point.
        Exit Sub
    End If
End If

End Sub

Public Function Euclid(ByRef Point1 As Point, ByRef Point2 As Point) As Double

EuclidsComputed = EuclidsComputed + 1
Euclid = Sqr(((Point2.X - Point1.X) * (Point2.X - Point1.X)) + ((Point2.Y - Point1.Y) * (Point2.Y - Point1.Y)))

End Function

Public Sub FindClosest(ByRef ThisPoint As Point, ByRef Tree() As Node)

Dim Closest As Long
Dim ThisEuclid As Double
Dim NodePointer As Long

MinEuclid = &H7FFFFFFF
FoundPoint = -1
EuclidsComputed = 0

NodePointer = GetNode(ThisPoint, Tree)
Call Ascend(ThisPoint, NodePointer, Tree)

End Sub

Public Function GetNode(ByRef ThisPoint As Point, ByRef Tree() As Node) As Long

Dim ThisNode As Long

Do
    If Tree(ThisNode).HasChildren Then 'Child is there. Iterate deeper. First, select the correct child.
        If ThisPoint.X >= Tree(ThisNode).Center.X And ThisPoint.Y < Tree(ThisNode).Center.Y Then 'NE
            If Not Tree(Tree(ThisNode).ChildNE).HasChildren Then
                If Tree(Tree(ThisNode).ChildNE).NumPoints < MAXOBJECTS Then
                    GetNode = Tree(ThisNode).ChildNE
                    Exit Function
                End If
            End If
            ThisNode = Tree(ThisNode).ChildNE
        ElseIf ThisPoint.X < Tree(ThisNode).Center.X And ThisPoint.Y < Tree(ThisNode).Center.Y Then 'NW
            If Not Tree(Tree(ThisNode).ChildNW).HasChildren Then
                If Tree(Tree(ThisNode).ChildNW).NumPoints < MAXOBJECTS Then
                    GetNode = Tree(ThisNode).ChildNW
                    Exit Function
                End If
            End If
            ThisNode = Tree(ThisNode).ChildNW
        ElseIf ThisPoint.X < Tree(ThisNode).Center.X And ThisPoint.Y >= Tree(ThisNode).Center.Y Then 'SW
            If Not Tree(Tree(ThisNode).ChildSW).HasChildren Then
                If Tree(Tree(ThisNode).ChildSW).NumPoints < MAXOBJECTS Then
                    GetNode = Tree(ThisNode).ChildSW
                    Exit Function
                End If
            End If
            ThisNode = Tree(ThisNode).ChildSW
        ElseIf ThisPoint.X >= Tree(ThisNode).Center.X And ThisPoint.Y >= Tree(ThisNode).Center.Y Then 'SE
            If Not Tree(Tree(ThisNode).ChildSE).HasChildren Then
                If Tree(Tree(ThisNode).ChildSE).NumPoints < MAXOBJECTS Then
                    GetNode = Tree(ThisNode).ChildSE
                    Exit Function
                End If
            End If
            ThisNode = Tree(ThisNode).ChildSE
        End If
    Else 'No children.
        GetNode = ThisNode
        Exit Function
    End If
Loop

End Function

Public Sub Ascend(ByRef ThisPoint As Point, ByVal ThisNode As Long, ByRef Tree() As Node)

Dim ThisEuclid As Double
Dim i As Long

'Call DrawGrid
'Form1.Picture1.Line (Tree(ThisNode).Bounds.Left, Tree(ThisNode).Bounds.Top)-(Tree(ThisNode).Bounds.Right, Tree(ThisNode).Bounds.Bottom), vbBlue, B

'Stop

If ThisPoint.Y - Tree(ThisNode).Bounds.Bottom > MinEuclid Then Exit Sub
If Tree(ThisNode).Bounds.Top - ThisPoint.Y > MinEuclid Then Exit Sub
If ThisPoint.X - Tree(ThisNode).Bounds.Right > MinEuclid Then Exit Sub
If Tree(ThisNode).Bounds.Left - ThisPoint.X > MinEuclid Then Exit Sub

If Tree(ThisNode).NumPoints > 0 Then
    For i = 0 To Tree(ThisNode).NumPoints - 1
        'Maybe apply margin test for points as well, instead of just for nodes.
        'If Abs(ThisPoint.X - Tree(ThisNode).MyPoints(i).X) < MinEuclid Or Abs(ThisPoint.Y - Tree(ThisNode).MyPoints(i).Y) < MinEuclid Then
            ThisEuclid = Euclid(ThisPoint, Tree(ThisNode).MyPoints(i))
            'Form1.Picture1.Circle (Tree(ThisNode).MyPoints(i).X, Tree(ThisNode).MyPoints(i).Y), 2, vbBlue
            'Stop
            If ThisEuclid < MinEuclid Then
                MinEuclid = ThisEuclid
                FoundPoint = Tree(ThisNode).MyIndex(i)
            End If
        'End If
    Next
End If

If Tree(ThisNode).HasChildren Then
    Call Descend(ThisPoint, Tree(ThisNode).ChildNE, Tree)
    Call Descend(ThisPoint, Tree(ThisNode).ChildNW, Tree)
    Call Descend(ThisPoint, Tree(ThisNode).ChildSW, Tree)
    Call Descend(ThisPoint, Tree(ThisNode).ChildSE, Tree)
End If

If ThisNode = 0 Then Exit Sub

Call Ascend(ThisPoint, Tree(ThisNode).Parent, Tree)

End Sub

Public Sub Descend(ByRef ThisPoint As Point, ByVal ThisNode As Long, ByRef Tree() As Node)

Dim ThisEuclid As Double
Dim i As Long

'Call DrawGrid
'Form1.Picture1.Line (Tree(ThisNode).Bounds.Left, Tree(ThisNode).Bounds.Top)-(Tree(ThisNode).Bounds.Right, Tree(ThisNode).Bounds.Bottom), vbBlue, B

'Stop

If ThisNode = 0 Then Exit Sub

If ThisPoint.Y - Tree(ThisNode).Bounds.Bottom > MinEuclid Then Exit Sub
If Tree(ThisNode).Bounds.Top - ThisPoint.Y > MinEuclid Then Exit Sub
If ThisPoint.X - Tree(ThisNode).Bounds.Right > MinEuclid Then Exit Sub
If Tree(ThisNode).Bounds.Left - ThisPoint.X > MinEuclid Then Exit Sub

If Tree(ThisNode).NumPoints > 0 Then
    For i = 0 To Tree(ThisNode).NumPoints - 1
        'Maybe apply margin test for points as well, instead of just for nodes.
        'If Abs(ThisPoint.X - Tree(ThisNode).MyPoints(i).X) < MinEuclid Or Abs(ThisPoint.Y - Tree(ThisNode).MyPoints(i).Y) < MinEuclid Then
            ThisEuclid = Euclid(ThisPoint, Tree(ThisNode).MyPoints(i))
            'Form1.Picture1.Circle (Tree(ThisNode).MyPoints(i).X, Tree(ThisNode).MyPoints(i).Y), 2, vbBlue
            'Stop
            If ThisEuclid < MinEuclid Then
                MinEuclid = ThisEuclid
                FoundPoint = Tree(ThisNode).MyIndex(i)
            End If
        'End If
    Next
End If

If Tree(ThisNode).HasChildren Then
    Call Descend(ThisPoint, Tree(ThisNode).ChildNE, Tree)
    Call Descend(ThisPoint, Tree(ThisNode).ChildNW, Tree)
    Call Descend(ThisPoint, Tree(ThisNode).ChildSW, Tree)
    Call Descend(ThisPoint, Tree(ThisNode).ChildSE, Tree)
End If

End Sub
